/* VzImagination (C) 2018-2019 Stephane Charette <stephanecharette@gmail.com>
 * $Id: TestImagination.cpp 2764 2019-04-30 21:50:34Z stephane $
 */

#include "TestUtil.hpp"
#include <tesseract/baseapi.h>


TEST(Imagination, DefaultConstructor)
{
	vz::Imagination img;
	ASSERT_TRUE(img.empty());
	ASSERT_TRUE(img.mats.empty());
	ASSERT_EQ(img.threshold_value, 0.0);
	ASSERT_EQ(img.threshold_type, vz::Imagination::EThresholdType::kInvalid);
	ASSERT_TRUE(img.contours.empty());
	ASSERT_TRUE(img.contours_intervals.empty());
}


TEST(Imagination, ThresholdWhiteness)
{
	std::map<std::string, int> m =
	{
		{ "capture"							, 23	},
		{ "coins"							, 9		},
		{ "sudoku"							, 29	},
		{ "sudoku_312"						, 15	},
		{ "page_29"							, 9		},
		{ "charter_of_rights_and_freedoms"	, 13	},
		{ "forrest_gump"					, 30	},
		{ "george_orwell"					, 12	},
		{ "mahatma_gandhi"					, 61	},
		{ "parliament_hill"					, 70	},
		{ "railroad"						, 63	},
		{ "the_foundry"						, 9		},
		{ "toronto_skyline"					, 51	},
		{ "ts_eliot"						, 8		},
		{ "wolfgang_amadeus_mozart"			, 49	}
	};

	// Create the threshold for the given image, and then look at the ratio between white and black pixels
	// compared to what we've historically seen.  When we enhance the threshold algorithm, we can expect
	// this value to change and require an update, but hopefully this will catch gross errors should we
	// completely break threshold functionality.

	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << name << std::endl;

		vz::Imagination img(path);
		cv::Mat original	= img.get(vz::Imagination::EMatType::kOriginalImage	);
		cv::Mat threshold	= img.get(vz::Imagination::EMatType::kThreshold		);

		cv::imwrite("whiteness_" + name + ".jpg", original, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
		cv::imwrite("whiteness_" + name + ".png", threshold, {cv::IMWRITE_PNG_COMPRESSION, 9});

		ASSERT_FALSE(threshold.empty());
		ASSERT_EQ(threshold.channels(), 1);
		ASSERT_EQ(threshold.rows, original.rows);
		ASSERT_EQ(threshold.cols, original.cols);

		const double white_count	= cv::countNonZero(threshold);
		const double total_count	= threshold.rows * threshold.cols;
		const int white_percentage	= std::round(100.0 * white_count / total_count);

		ASSERT_EQ(white_percentage, m[name]);
	}
}


TEST(Imagination, ErosionIterations)
{
	vz::VMats v;

	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << name << std::endl;

		vz::Imagination img(path);

		for (int number_of_iterations : {1, 2, 3, 5, 9})
		{
			std::cout << name << ": erosion iteration #" << number_of_iterations << std::endl;

			cv::Mat dst = img.get(vz::Imagination::EMatType::kErode, number_of_iterations);
			cv::imwrite("erosion_" + name + "_iteration_" + std::to_string(number_of_iterations) + ".jpg", dst, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
			ASSERT_FALSE(dst.empty());
			ASSERT_EQ(dst.channels(), 1);

			// check that this erosion is different from all the other eroded images
			std::cout << name << ": comparing eroded image against " << v.size() << " previous images" << std::endl;
			for (auto & m : v)
			{
				ASSERT_FALSE(vz::compare_exact(m, dst));
			}

			// don't bother remembering some of the images, once eroded they're pure black and we end up with duplicates
			if (name != "sudoku" &&
				name != "sudoku_312" &&
				name != "page_29" &&
				name != "mahatma_gandhi" &&
				name != "railroad" &&
				name != "toronto_skyline" &&
				name != "ts_eliot" &&
				name != "mahatma_gandhi" &&
				name != "wolfgang_amadeus_mozart")
			{
				v.push_back(dst);
			}
		}
	}
}


TEST(Imagination, ErosionDilationIterations)
{
	vz::VMats v;

	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << name << std::endl;

		vz::Imagination img(path);

		for (int number_of_iterations : {1, 2, 3, 5, 9})
		{
			std::cout << name << ": erosion and dilation iteration #" << number_of_iterations << std::endl;

			cv::Mat dst = img.get(vz::Imagination::EMatType::kErodeAndDilate, number_of_iterations);
			ASSERT_EQ(dst.channels(), 1);
			ASSERT_FALSE(dst.empty());

			// check that this erosion + dilation is different from all the other images
			std::cout << name << ": comparing erosion+dilation image against " << v.size() << " previous images" << std::endl;
			for (cv::Mat mat : v)
			{
				ASSERT_FALSE(vz::compare_exact(mat, dst));
			}

			// don't bother remembering some of the images, once eroded they're pure black and we end up with duplicates
			if (name != "sudoku" &&
				name != "sudoku_312" &&
				name != "page_29" &&
				name != "mahatma_gandhi" &&
				name != "railroad" &&
				name != "toronto_skyline" &&
				name != "ts_eliot" &&
				name != "mahatma_gandhi" &&
				name != "wolfgang_amadeus_mozart")
			{
				v.push_back(dst);
			}
		}
	}
}


TEST(Imagination, SaveAll)
{
	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << "save all: " << name << std::endl;
		vz::Imagination img(path);

		auto v = img.save_all("save_" + name + "_");
		ASSERT_EQ(v.size(), static_cast<size_t>(vz::Imagination::EMatType::kMax) - 2); // -2 because we skip over kInvalid and kUserDefined

		std::cout << name << ": threshold=" << img.threshold_value << std::endl;
		ASSERT_NE(0.0, img.threshold_value);

		std::cout << name << ": contours=" << img.contours.size() << std::endl;
		ASSERT_FALSE(img.contours.empty());
		ASSERT_NE(0, img.contours.size());

		std::cout << name << ": contour intervals=" << img.contours_intervals.size() << std::endl;
//		ASSERT_FALSE(img.contours_intervals.empty());
//		ASSERT_NE(0, img.contours_intervals.size());

		std::cout << name << ": hough lines=" << img.hough_lines.size() << std::endl;
		ASSERT_FALSE(img.hough_lines.empty());
		ASSERT_NE(0, img.hough_lines.size());

		std::cout << name << ": hough lines angles=" << img.hough_lines_angles.size() << std::endl;
		ASSERT_FALSE(img.hough_lines_angles.empty());
		ASSERT_NE(0, img.hough_lines_angles.size());
		ASSERT_EQ(img.hough_lines.size(), img.hough_lines_angles.size());

		std::cout << name << ": hough lines intervals=" << img.hough_lines_angle_intervals.size() << std::endl;
		ASSERT_FALSE(img.hough_lines_angle_intervals.empty());
		ASSERT_NE(0, img.hough_lines_angle_intervals.size());

		std::cout << name << ": hough lines correction angle=" << img.hough_lines_correction_angle << std::endl;
		ASSERT_NE(0.0, img.hough_lines_correction_angle);

		// now go through all the saved files and make sure they can be loaded
		for (auto filename : v)
		{
			std::cout << "saved file " << filename << std::endl;

			vz::Imagination test(filename);
			ASSERT_FALSE(test.empty());
		}
	}
}


TEST(Imagination, FindContours)
{
	for (const auto & [path, name] : complex_test_images())
	{
		std::cout << name << std::endl;

		vz::Imagination img(path);

		cv::Mat mat;
		mat = img.get(vz::Imagination::EMatType::kFindContours);
		cv::imwrite("contours_" + name + ".jpg", mat, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		mat = img.get(vz::Imagination::EMatType::kFindContoursIntervals);
		cv::imwrite("contours_intervals_" + name + ".jpg", mat, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});

		vz::VDouble data;
		for (auto c : img.contours)
		{
			const double area = cv::contourArea(c);
			data.push_back(area);
		}
		vz::VVDouble vv = vz::intervals(data, 3.5);
		std::cout << name << " " << vz::to_string(vv) << std::endl;
	}
}


TEST(Imagination, HoughLines)
{
	for (const auto & [path, name] : complex_test_images())
	{
		vz::Imagination img(path);
		cv::Mat canny = img.get(vz::Imagination::EMatType::kCannyEdgeDetection);

		vz::VV4i lines;
		cv::HoughLinesP(
				canny,
				lines,
				1,			// distance resolution of the accumulator in pixels
				CV_PI/360,	// angle resolution of the accumulator in radians
				80,			// accumulator threshold parameter
				5,			// minimum line length; line segments shorter than that are rejected
				10);		// maximum allowed gap between points on the same line to link them

		// draw the lines onto the original image
		cv::Mat dst = img.get(vz::Imagination::EMatType::kO2SimpleColourBalance);
		const int thickness = 1;
		const int line_type = cv::LINE_AA;
		for(size_t i = 0; i < lines.size(); i++)
		{
			cv::line(dst,
					cv::Point(lines[i][0], lines[i][1]),
					cv::Point(lines[i][2], lines[i][3]),
					cv::Scalar(0,0,255),
					thickness, line_type);
		}

		cv::imwrite("hough_lines_" + name + ".jpg", dst, {cv::IMWRITE_JPEG_QUALITY, 75, cv::IMWRITE_JPEG_OPTIMIZE, 1});
	}
}


TEST(Imagination, OCR)
{
	test_images();

	vz::Imagination img("page_29.jpg");
//	vz::Imagination img("page_29.png");
//	vz::Imagination img("sudoku_312.jpg");
	cv::Mat mat = img.get(vz::Imagination::EMatType::kO2SimpleColourBalance);

	tesseract::TessBaseAPI tess;
	tess.Init(nullptr, "eng", tesseract::OEM_DEFAULT);
	tess.SetImage(mat.data, mat.cols, mat.rows, mat.channels(), mat.step1());

	char *text = tess.GetUTF8Text();
	std::cout << "TEXT: " << text << std::endl;
	delete [] text;

	tess.End();
}
